import contextlib

from . import constants, keys
from .client import S3Client
from .configs import S3FileTypeConfig

with contextlib.suppress(ImportError):
    from .async_client import AsyncS3Client

with contextlib.suppress(ImportError):
    from . import testing

with contextlib.suppress(ImportError):
    from . import factory

__all__ = (
    "AsyncS3Client",
    "S3Client",
    "S3FileTypeConfig",
    "constants",
    "factory",
    "keys",
    "testing",
)



import dataclasses
import typing

from . import keys


class S3FileTypeConfigMeta(type):
    """Meta class for S3FileTypeConfig."""

    def __call__(
        cls,
        *args,  # noqa: ANN002
        **kwargs,
    ) -> "S3FileTypeConfig":
        """Update mapping of S3SupportedFieldConfigs."""
        instance: S3FileTypeConfig = super().__call__(*args, **kwargs)
        if instance.name in S3FileTypeConfig.configs:
            raise ValueError(f"{instance.name} config is already defined")
        S3FileTypeConfig.configs[instance.name] = instance
        return instance


@dataclasses.dataclass(frozen=True)
class S3FileTypeConfig(metaclass=S3FileTypeConfigMeta):
    """Configuration for S3 file upload."""

    configs: typing.ClassVar[dict[str, "S3FileTypeConfig"]] = {}

    name: str
    # S3Key are used to generate file's path
    key: keys.S3Key
    # Mime types are allowed, None - for all
    allowed: tuple[str, ...] | None = None
    # Perform checks against user
    auth: typing.Callable[[typing.Any | None], bool] | None = None
    # Define allowed size limits for file (in bytes)
    content_length_range: tuple[int, int] | None = None
    # In how much second pre-signed URL for upload will expire
    expires_in: int = 3600
    success_action_status: int = 201
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Disposition
    content_disposition: (
        typing.Literal["attachment"] | typing.Literal["inline"]
    ) = "attachment"



from django.core.files.storage import default_storage

from .. import client


def get_s3_client() -> client.S3Client:
    """Get s3 client based on Django default storage."""
    return client.S3Client(
        boto3_client=default_storage.connection.meta.client,  # type: ignore
        default_bucket=default_storage.bucket_name,  # type: ignore
    )



import contextlib
import dataclasses
import typing

from rest_framework import (
    decorators,
    exceptions,
    permissions,
    response,
    status,
    viewsets,
)
from rest_framework.request import Request

from .. import client, configs
from . import serializers, shortcuts


class S3GetParamsView(viewsets.GenericViewSet):
    """View for getting params for s3 to upload file to S3."""

    serializer_class = serializers.S3RequestParamsSerializer
    permission_classes = (permissions.AllowAny,)
    filter_backends = ()
    pagination_class = None

    @decorators.action(
        methods=["POST"],
        url_path="get-params",
        url_name="get-params",
        detail=False,
    )
    def get_params(
        self,
        request: Request,
    ) -> response.Response:
        """Get parameters for upload to S3 bucket.

        Current endpoint returns all required for s3 upload data,
        which should be later sent to `url` as `form-data` url with
        'file'. Workflow: First, you make request to this endpoint. Then send
        response data to `url` via `POST` as form-data with file included. In
        response you will get an url which you can use in API for value of file
        related fields like avatar for example.

        """
        serializer = self.serializer_class(
            context_request=request,
            data=request.data,
        )
        serializer.is_valid(raise_exception=True)
        s3_client = self.get_s3_client()
        params = s3_client.generate_params(
            filename=serializer.data["filename"],  # type: ignore
            config=serializer.data["config"],  # type: ignore
            content_type=serializer.data["content_type"],  # type: ignore
            extra_metadata=self.get_extra_meta_data(user=request.user),
        )
        return response.Response(
            status=status.HTTP_200_OK,
            data=serializers.S3UploadSerializer(
                instance=dataclasses.asdict(params),
            ).data,
        )

    @decorators.action(
        methods=["GET"],
        url_path="list-configs",
        url_name="list-configs",
        detail=False,
    )
    def list_configs(
        self,
        request: Request,
    ) -> response.Response:
        """List all configs for s3 upload."""
        return response.Response(
            status=status.HTTP_200_OK,
            data=serializers.S3ConfigSerializer(
                instance=map(
                    dataclasses.asdict,
                    configs.S3FileTypeConfig.configs.values(),
                ),
                many=True,
            ).data,
        )

    @decorators.action(
        methods=["GET"],
        url_path="retrieve-config/(?P<name>[^/.]+)",
        url_name="retrieve-config",
        detail=False,
    )
    def retrieve_config(
        self,
        request: Request,
        name: str,
    ) -> response.Response:
        """Retrieve config for s3 upload."""
        if name not in configs.S3FileTypeConfig.configs:
            raise exceptions.NotFound
        return response.Response(
            status=status.HTTP_200_OK,
            data=serializers.S3ConfigSerializer(
                instance=dataclasses.asdict(
                    configs.S3FileTypeConfig.configs[name],
                ),
            ).data,
        )

    def get_s3_client(self) -> client.S3Client:
        """Get s3 client for params generation."""
        return shortcuts.get_s3_client()

    def get_extra_meta_data(
        self,
        user: typing.Any,
    ) -> dict[str, str]:
        """Extend meta data for file."""
        return {
            "user-id": str(user.pk),
        }


with contextlib.suppress(ImportError):
    import drf_spectacular.utils

    drf_spectacular.utils.extend_schema_view(
        get_params=drf_spectacular.utils.extend_schema(
            request=serializers.S3RequestParamsSerializer,
            responses=serializers.S3UploadSerializer,
        ),
        list_configs=drf_spectacular.utils.extend_schema(
            responses=serializers.S3ConfigSerializer(many=True),
        ),
        retrieve_config=drf_spectacular.utils.extend_schema(
            parameters=[
                drf_spectacular.utils.OpenApiParameter(
                    name="name",
                    type=str,
                    required=True,
                    location=drf_spectacular.utils.OpenApiParameter.PATH,
                ),
            ],
            responses=serializers.S3ConfigSerializer(),
        ),
    )(S3GetParamsView)



import collections.abc
import io
import random

import factory
import PIL.Image
import PIL.ImageColor

import mypy_boto3_s3

from . import client, configs

BucketGetter = collections.abc.Callable[
    [],
    str,
]


class S3FileField(factory.LazyAttribute):
    """Generate file and upload to s3."""

    _boto3_client: mypy_boto3_s3.S3Client | None = None

    def __init__(
        self,
        s3_config: str,
        s3_region: client.RegionGetter | str,
        bucket: BucketGetter | str,
        access_key_getter: client.AccessKeyGetter,
        s3_endpoint_url_getter: client.S3EndpointUrlGetter | None = None,
        filename: str = "example.txt",
        data: bytes = b"Test",
    ) -> None:
        self.access_key_getter = access_key_getter
        self.s3_endpoint_url_getter = s3_endpoint_url_getter
        self.bucket = bucket
        self.s3_region = s3_region
        self.filename = filename
        self.s3_config = s3_config
        self.data = data

        super().__init__(
            function=self._generate,
        )

    def _get_boto3(self) -> mypy_boto3_s3.S3Client:
        if not self._boto3_client:
            self._boto3_client = client.get_boto3_s3_client(
                region=self.s3_region,
                s3_endpoint_url_getter=self.s3_endpoint_url_getter,
                access_key_getter=self.access_key_getter,
            )
        return self._boto3_client

    def _get_s3_client(self) -> client.S3Client:
        """Set up s3 client."""
        if isinstance(self.bucket, str):
            bucket = self.bucket
        if callable(self.bucket):
            bucket = self.bucket()
        return client.S3Client(
            boto3_client=self._get_boto3(),
            default_bucket=bucket,
        )

    def _generate_file_data(self) -> bytes:
        """Generate data for file."""
        return self.data

    def _generate(
        self,
        *args,  # noqa: ANN002
        **kwargs,
    ) -> str:
        """Generate file and upload it to s3."""
        return self._get_s3_client().upload_file(
            filename=self.filename,
            file_obj=io.BytesIO(self._generate_file_data()),
            config=configs.S3FileTypeConfig.configs[self.s3_config],
        )


class S3ImageFileField(S3FileField):
    """Generate image file and upload to s3."""

    def __init__(
        self,
        s3_config: str,
        s3_region: str,
        bucket: str,
        access_key_getter: client.AccessKeyGetter,
        s3_endpoint_url_getter: client.S3EndpointUrlGetter | None = None,
        filename: str = "example.jpg",
        data: bytes = b"",
        color: str = "",
        width: int = 0,
        height: int = 0,
    ) -> None:
        self.color = color
        self.width = width
        self.height = height
        super().__init__(
            s3_config=s3_config,
            s3_region=s3_region,
            bucket=bucket,
            access_key_getter=access_key_getter,
            s3_endpoint_url_getter=s3_endpoint_url_getter,
            data=data,
            filename=filename,
        )

    def _generate_file_data(self) -> bytes:
        """Generate image data for file."""
        width = self.width or self.height or 100
        height = self.height or self.width or 100
        color = self.color or random.choice(  # noqa: S311
            list(PIL.ImageColor.colormap.keys()),
        )

        thumb_io = io.BytesIO()
        with PIL.Image.new("RGB", (width, height), color) as thumb:
            thumb.save(thumb_io, format="JPEG")
        return thumb_io.getvalue()



import typing

import humanize
from django.conf import settings
from django.db import models
from rest_framework import exceptions, request, serializers

from .. import configs, constants
from . import drf_fields


class S3FieldsConfigMixin:
    """Extend serializer_field_mapping with s3 fields.

    Deprecate it once encode/django-rest-framework#9507 is accepted.

    """

    @property
    def serializer_field_mapping(
        self,
    ) -> dict[type[models.Field], type[serializers.Field]]:
        """Extend serializer mapping with s3 fields."""
        serializer_field_mapping = super().serializer_field_mapping  # type: ignore
        serializer_field_mapping[models.FileField] = (
            drf_fields.S3UploadURLField
        )
        serializer_field_mapping[models.ImageField] = (
            drf_fields.S3UploadURLField
        )
        return serializer_field_mapping


class S3RequestParamsSerializer(serializers.Serializer):
    """Serializer for validation s3 uploading fields."""

    config = drf_fields.S3FileTypeConfigField()
    filename = serializers.CharField()
    content_type = serializers.CharField()
    content_length = serializers.IntegerField()

    def __init__(
        self,
        context_request: request.Request | None = None,
        *args,  # noqa: ANN002
        **kwargs,
    ) -> None:
        """Set current user."""
        super().__init__(*args, **kwargs)
        self._request: request.Request | None = context_request
        self._user = getattr(self._request, "user", None)

    def validate_config(
        self,
        config: configs.S3FileTypeConfig,
    ) -> configs.S3FileTypeConfig:
        """Check that user can use dest."""
        if config.auth and not config.auth(self._user):
            raise exceptions.ValidationError(
                "Current user can't use this destination",
            )
        return config

    def validate(self, attrs: dict[str, typing.Any]) -> dict[str, typing.Any]:
        """Perform validations.

        Check what input data is valid for specified configuration.

        """
        errors: dict[str, str] = {}
        config: configs.S3FileTypeConfig = attrs["config"]
        filename: str = attrs["filename"]
        content_type: str = attrs["content_type"]
        content_length: int = attrs["content_length"]
        if config.allowed and content_type not in config.allowed:
            expected = ", ".join(config.allowed)
            errors["content_type"] = (
                f"Invalid file type - `{content_type}` of `{filename}`. "
                f"Expected: {expected}."
            )
        if config.content_length_range:
            min_bound, max_bound = config.content_length_range
            if min_bound > content_length:
                errors["content_length"] = (
                    "Invalid file size "
                    f"- {humanize.naturalsize(content_length)} "
                    f"of {filename}. "
                    f"Need between {humanize.naturalsize(min_bound)} "
                    f"and {humanize.naturalsize(max_bound)}."
                )
            if max_bound < content_length:
                errors["content_length"] = (
                    "Invalid file size "
                    f"- {humanize.naturalsize(content_length)} "
                    f"of {filename}. "
                    f"Need between {humanize.naturalsize(min_bound)} "
                    f"and {humanize.naturalsize(max_bound)}."
                )
        if errors:
            raise exceptions.ValidationError(errors)
        return attrs


class S3ParamsSerializer(serializers.Serializer):
    """Serializer for showing params for s3 upload."""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        fields = getattr(
            settings,
            "SARITASA_S3_TOOLS_UPLOAD_PARAMS",
            constants.s3v4_signature_fields,
        )
        for field in fields:
            self.fields[field] = serializers.CharField(
                label=field,
                required=False,
                allow_null=True,
            )


class S3UploadSerializer(serializers.Serializer):
    """Serializer auto swagger documentation.

    This serializer used just for packages that capable to generate
    openapi/swagger specs, so that front-end team
    could see specs for response for view.

    """

    url = serializers.URLField()
    params = S3ParamsSerializer()


class S3ConfigSerializer(serializers.Serializer):
    """Serializer to represent s3 config."""

    name = serializers.CharField()
    allowed = serializers.ListField(
        child=serializers.CharField(),
        allow_null=True,
    )
    content_length_range = serializers.ListField(
        child=serializers.CharField(),
        allow_null=True,
    )
    expires_in = serializers.IntegerField()
    success_action_status = serializers.IntegerField()
    content_disposition = serializers.CharField()



import collections.abc
import contextlib
import typing

import pytest

import boto3
import botocore.config
import botocore.credentials
import botocore.exceptions
import mypy_boto3_s3
import mypy_boto3_s3.type_defs

import saritasa_s3_tools


def pytest_addoption(parser: pytest.Parser) -> None:
    """Set up cmd args."""
    parser.addini(
        "s3_access_key",
        "Access key for s3.",
        default="",
    )
    parser.addini(
        "s3_secret_key",
        "Secret key for s3.",
        default="",
    )
    parser.addini(
        "s3_endpoint_url",
        "Endpoint for s3.",
        default="",
    )
    parser.addini(
        "s3_region",
        "Region for s3.",
        default="",
    )
    parser.addini(
        "s3_bucket_name",
        "Bucket for s3.",
        default="saritasa-s3-tools",
    )


@pytest.fixture(scope="session")
def access_key_getter(
    request: pytest.FixtureRequest,
) -> collections.abc.Callable[
    [],
    botocore.credentials.Credentials,
]:
    """Set up cred getter."""
    if (
        s3_access_key := request.config.getini(
            "s3_access_key",
        )
    ) and (
        s3_secret_key := request.config.getini(
            "s3_secret_key",
        )
    ):
        return lambda: botocore.credentials.Credentials(
            access_key=str(s3_access_key),
            secret_key=str(s3_secret_key),
        )
    raise NotImplementedError(  # pragma: no cover
        "Please set up `access_key_getter` fixture or "
        "set `s3_access_key` and `s3_secret_key` in `.ini` file.",
    )


@pytest.fixture(scope="session")
def s3_endpoint_url_getter(
    request: pytest.FixtureRequest,
) -> (
    collections.abc.Callable[
        [],
        str | None,
    ]
    | None
):
    """Set up url getter."""
    if s3_endpoint_url := request.config.getini("s3_endpoint_url"):
        return lambda: str(s3_endpoint_url)
    return None  # pragma: no cover


@pytest.fixture(scope="session")
def s3_region(
    request: pytest.FixtureRequest,
) -> str:
    """Get s3 region."""
    return str(request.config.getini("s3_region"))


@pytest.fixture(scope="session")
def aws_session(
    access_key_getter: collections.abc.Callable[
        [],
        botocore.credentials.Credentials,
    ],
    s3_region: str,
) -> boto3.session.Session:
    """Get aws session."""
    return saritasa_s3_tools.client.get_boto3_session(
        access_key_getter=access_key_getter,
        region=s3_region,
    )


@pytest.fixture(scope="session")
def aws_config() -> botocore.config.Config | None:
    """Get aws config."""
    return None


@pytest.fixture(scope="session")
def boto3_resource(
    request: pytest.FixtureRequest,
) -> mypy_boto3_s3.S3ServiceResource:
    """Prepare boto3 resource."""
    try:
        return request.getfixturevalue("boto3_resource_from_django")
    except ImportError:  # pragma: no cover
        return request.getfixturevalue("boto3_resource_from_config")


@pytest.fixture(scope="session")
def boto3_resource_from_django() -> mypy_boto3_s3.S3ServiceResource:
    """Get boto3 resource for django storage."""
    from django.core.files.storage import default_storage

    return default_storage.connection  # type: ignore


@pytest.fixture(scope="session")
def boto3_resource_from_config(
    aws_session: boto3.session.Session,
    aws_config: botocore.config.Config,
    s3_endpoint_url_getter: collections.abc.Callable[
        [],
        str | None,
    ]
    | None,
) -> mypy_boto3_s3.S3ServiceResource:
    """Prepare boto3 resource."""
    return saritasa_s3_tools.client.get_boto3_s3_resource(
        session=aws_session,
        s3_endpoint_url_getter=s3_endpoint_url_getter,
        config=aws_config,
    )


@pytest.fixture(scope="session")
def boto3_client(
    boto3_resource: mypy_boto3_s3.S3ServiceResource,
) -> mypy_boto3_s3.S3Client:
    """Prepare boto3 client."""
    return boto3_resource.meta.client


@pytest.fixture(scope="session")
def s3_bucket_name(
    request: pytest.FixtureRequest,
) -> str:
    """Get the name of s3 bucket."""
    worker_input = getattr(
        request.config,
        "workerinput",
        {
            "workerid": "",
        },
    )
    return "-".join(
        filter(
            None,
            (
                str(request.config.getini("s3_bucket_name")),
                worker_input["workerid"],
            ),
        ),
    )


@pytest.fixture(scope="session")
def s3_bucket_cleaner(
    boto3_resource: mypy_boto3_s3.ServiceResource,
) -> collections.abc.Callable[[str], None]:
    """Get bucket cleaner."""

    def _clean(bucket: str) -> None:
        boto3_resource.Bucket(bucket).objects.all().delete()

    return _clean


@pytest.fixture(scope="session")
def s3_bucket_factory(
    boto3_client: mypy_boto3_s3.S3Client,
    s3_region: str,
    s3_bucket_cleaner: collections.abc.Callable[[str], None],
) -> collections.abc.Callable[
    [str],
    contextlib._GeneratorContextManager[
        mypy_boto3_s3.type_defs.CreateBucketOutputTypeDef
    ],
]:
    """Get factory for generating s3 buckets."""

    @contextlib.contextmanager
    def _create(
        bucket: str,
    ) -> collections.abc.Generator[
        mypy_boto3_s3.type_defs.CreateBucketOutputTypeDef,
        None,
        None,
    ]:
        with contextlib.suppress(botocore.exceptions.ClientError):
            boto3_client.head_bucket(Bucket=bucket)
            s3_bucket_cleaner(bucket)  # pragma: no cover
            boto3_client.delete_bucket(Bucket=bucket)  # pragma: no cover
        yield boto3_client.create_bucket(
            Bucket=bucket,
            CreateBucketConfiguration={
                "LocationConstraint": s3_region,  # type: ignore
            },
        )
        s3_bucket_cleaner(bucket)
        boto3_client.delete_bucket(Bucket=bucket)

    return _create


@pytest.fixture(scope="session")
def s3_bucket(
    s3_bucket_factory: collections.abc.Callable[
        [str],
        contextlib._GeneratorContextManager[
            mypy_boto3_s3.type_defs.CreateBucketOutputTypeDef
        ],
    ],
    s3_bucket_name: str,
) -> collections.abc.Generator[
    str,
    None,
    None,
]:
    """Create s3 bucket."""
    with s3_bucket_factory(s3_bucket_name) as _:
        yield s3_bucket_name


@pytest.fixture(scope="session")
def s3_client(
    boto3_client: mypy_boto3_s3.S3Client,
    s3_bucket: str,
) -> saritasa_s3_tools.S3Client:
    """Set up s3 client."""
    return saritasa_s3_tools.S3Client(
        boto3_client=boto3_client,
        default_bucket=s3_bucket,
    )


@pytest.fixture(scope="session")
def async_s3_client(
    boto3_client: mypy_boto3_s3.S3Client,
    s3_bucket: str,
) -> saritasa_s3_tools.AsyncS3Client:
    """Set up async s3 client."""
    return saritasa_s3_tools.AsyncS3Client(
        boto3_client=boto3_client,
        default_bucket=s3_bucket,
    )


@pytest.fixture
def django_storage_changer() -> collections.abc.Iterator[
    collections.abc.Callable[
        [str, typing.Any],
        None,
    ]
]:
    """Temporary change default storage settings."""
    from django.core.files.storage import default_storage

    old_settings: dict[str, typing.Any] = {}

    def _changer(key: str, value: typing.Any) -> None:
        if key not in old_settings and hasattr(default_storage, key):
            old_settings[key] = getattr(default_storage, key)
        setattr(default_storage, key, value)

    yield _changer
    for key, value in old_settings.items():
        setattr(default_storage, key, value)


@pytest.fixture(scope="session")
def django_adjust_s3_bucket(s3_bucket: str) -> None:
    """Set bucket to a test one."""
    from django.conf import settings
    from django.core.files import storage
    from storages.backends import s3

    for storage_alias in settings.STORAGES:
        with contextlib.suppress(storage.InvalidStorageError):
            storage_instance = storage.storages[storage_alias]
            if isinstance(storage_instance, s3.S3Storage):
                storage_instance.bucket_name = s3_bucket  # type: ignore
    settings.AWS_STORAGE_BUCKET_NAME = s3_bucket
    return None



import collections.abc
import functools
import typing

import anyio

import mypy_boto3_s3.type_defs

from .. import client, configs

ReturnT = typing.TypeVar("ReturnT")
ParamT = typing.ParamSpec("ParamT")


class AsyncS3Client(client.S3Client):
    """Async Client for interacting with s3 based on boto3 client."""

    async def run_sync_as_async(
        self,
        func: collections.abc.Callable[ParamT, ReturnT],
        *args: ParamT.args,
        **kwargs: ParamT.kwargs,
    ) -> ReturnT:
        """Make sync function run in async env."""
        return await anyio.to_thread.run_sync(  # type: ignore
            functools.partial(func, *args, **kwargs),
        )

    async def async_generate_params(
        self,
        filename: str,
        config: configs.S3FileTypeConfig,
        content_type: str,
        bucket: str = "",
        upload_folder: str = "",
        extra_metadata: dict[str, str] | None = None,
    ) -> client.S3UploadParams:
        """Generate params for s3 upload in async env."""
        return await self.run_sync_as_async(
            self.generate_params,
            filename=filename,
            upload_folder=upload_folder,
            config=config,
            bucket=bucket,
            content_type=content_type,
            extra_metadata=extra_metadata,
        )

    async def async_upload_file(
        self,
        filename: str,
        config: configs.S3FileTypeConfig,
        file_obj: mypy_boto3_s3.type_defs.FileobjTypeDef,
        bucket: str = "",
    ) -> str:
        """Upload file to s3 in async env."""
        return await self.run_sync_as_async(
            self.upload_file,
            filename=filename,
            config=config,
            bucket=bucket,
            file_obj=file_obj,
        )

    async def async_download_file(
        self,
        key: str,
        file_obj: mypy_boto3_s3.type_defs.FileobjTypeDef,
        bucket: str = "",
    ) -> mypy_boto3_s3.type_defs.FileobjTypeDef:
        """Download file from s3 in async env."""
        return await self.run_sync_as_async(
            self.download_file,
            file_obj=file_obj,
            bucket=bucket,
            key=key,
        )

    async def async_get_file_metadata(
        self,
        key: str,
        bucket: str = "",
    ) -> mypy_boto3_s3.type_defs.HeadObjectOutputTypeDef:
        """Get file's metadata in async env."""
        return await self.run_sync_as_async(
            self.get_file_metadata,
            bucket=bucket,
            key=key,
        )

    async def async_is_file_in_bucket(
        self,
        key: str,
        bucket: str = "",
    ) -> bool:
        """Check if file is in bucket in async env."""
        return await self.run_sync_as_async(
            self.is_file_in_bucket,
            bucket=bucket,
            key=key,
        )

    async def async_copy_object(
        self,
        key: str,
        source_key: str,
        bucket: str = "",
        source_bucket: str = "",
    ) -> None:
        """Copy file object from copy source to key path in async env."""
        return await self.run_sync_as_async(
            self.copy_object,
            key=key,
            source_key=source_key,
            bucket=bucket,
            source_bucket=source_bucket,
        )

    async def async_delete_object(
        self,
        key: str,
        bucket: str = "",
    ) -> None:
        """Delete file object from s3 bucket is async env."""
        return await self.run_sync_as_async(
            self.delete_object,
            key=key,
            bucket=bucket,
        )



from .drf_fields import S3FileTypeConfigField, S3UploadURLField
from .model_fields import S3FileField, S3FileFieldMixin, S3ImageField
from .serializers import (
    S3ConfigSerializer,
    S3FieldsConfigMixin,
    S3ParamsSerializer,
    S3RequestParamsSerializer,
    S3UploadSerializer,
)
from .shortcuts import get_s3_client
from .views import S3GetParamsView



import pathlib
import xml.etree.ElementTree

import httpx

from .. import client


def upload_file(
    filepath: str,
    s3_params: client.S3UploadParams,
) -> httpx.Response:
    """Upload file to s3."""
    url = s3_params.url
    params = s3_params.params
    # Test file upload itself
    with (
        httpx.Client() as client,
        pathlib.Path(filepath).open("rb") as upload_file,
    ):
        upload_response = client.post(
            url=url,
            data={
                key: value
                for key, value in params.items()
                if value is not None
            },
            files={"file": upload_file.read()},
        )
    return upload_response


def upload_file_and_verify(
    filepath: str,
    s3_params: client.S3UploadParams,
    is_minio: bool = False,
) -> tuple[str, str]:
    """Upload and verify that file is uploaded."""
    upload_response = upload_file(
        filepath=filepath,
        s3_params=s3_params,
    )
    assert upload_response.is_success, upload_response.content  # noqa: S101
    parsed_response = xml.etree.ElementTree.fromstring(  # noqa: S314
        upload_response.content.decode(),
    )
    if is_minio:
        file_key = parsed_response[1].text
        file_url = parsed_response[3].text
    else:
        file_key = parsed_response[2].text
        file_url = parsed_response[0].text
    assert file_url, upload_response.content  # noqa: S101
    assert file_key, upload_response.content  # noqa: S101
    return file_url, file_key



import abc
import pathlib
import re
import unicodedata
import uuid


class S3Key:
    """Base class for s3 keys."""

    uuid_regex = r"[\d|\w]{8}-[\d|\w]{4}-[\d|\w]{4}-[\d|\w]{4}-[\d|\w]{12}"

    @abc.abstractmethod
    def __call__(self, filename: str | None) -> str:
        """Abstract method for calling keys."""

    def remove_special_characters(self, filename: str) -> str:
        """Remove characters from filename that are not allowed in some OS."""
        special_characters = r"<>:\"/\\|?*"
        return filename.translate({ord(i): None for i in special_characters})

    def normalize_string_value(self, value: str) -> str:
        """Normalize string value.

        1. Remove leading and trailing whitespaces.
        2. Replace all space characters with the Space char.
        3. Normalize Unicode string using `NFKC` form. See the details:
        https://docs.python.org/3/library/unicodedata.html#unicodedata.normalize

        """
        cleaned = " ".join(value.strip().split()).strip()
        return unicodedata.normalize("NFKC", cleaned)

    def clean_filename(self, filename: str) -> str:
        """Remove `garbage` characters that cause problems with file names."""
        cleaned = self.remove_special_characters(filename)
        normalized = self.normalize_string_value(cleaned)

        return normalized

    def get_random_filename(self, filename: str) -> str:
        """Get random filename.

        Generation random filename that contains unique identifier and
        filename extension like: ``photo.jpg``.

        Args:
        ----
            filename (str): Name of file.

        Returns:
        -------
            new_filename (str): ``9841422d-c041-45a5-b7b3-467179f4f127.ext``.

        """
        path = str(uuid.uuid4())
        ext = pathlib.Path(filename).suffix.lower()

        return "".join((path, ext))

    def validate(self, key: str) -> bool:
        """Check that input key is matching Key pattern."""
        return True  # pragma: no cover


class WithPrefixUUIDFileName(S3Key):
    """Generate S3 key with prefix folder and uuid filename.

    Example:
    -------
        prefix/{UUID.extension}

    """

    def __init__(self, prefix: str) -> None:
        self.prefix = prefix.removesuffix("/")

    def __call__(self, filename: str | None) -> str:
        """Return prefixed S3 key."""
        if not filename:
            return f"{self.prefix}/{uuid.uuid4()}.incorrect"
        return f"{self.prefix}/{self.get_random_filename(filename)}"

    def validate(self, key: str) -> bool:
        """Check that input key is matching Key pattern."""
        return bool(
            re.compile(pattern=rf"{self.prefix}/{self.uuid_regex}\..+").match(
                key,
            ),
        )


class WithPrefixUUIDFolder(S3Key):
    """Generate S3 key with prefix folder and uuid folder.

    Example:
    -------
        prefix/{UUID}/filename

    """

    def __init__(self, prefix: str) -> None:
        self.prefix = prefix.removesuffix("/")

    def __call__(self, filename: str | None) -> str:
        """Create key for destination using filename."""
        if not filename:
            return f"{self.prefix}/{uuid.uuid4()}/{uuid.uuid4()}.incorrect"
        return f"{self.prefix}/{uuid.uuid4()}/{self.clean_filename(filename)}"

    def validate(self, key: str) -> bool:
        """Check that input key is matching Key pattern."""
        return bool(
            re.compile(
                pattern=rf"{self.prefix}/{self.uuid_regex}/.+\..+",
            ).match(
                key,
            ),
        )



import collections.abc

from django.core import exceptions
from django.core.files import utils
from django.db import models
from django.db.models.fields import files
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

import botocore.exceptions

from .. import configs


class S3FileFieldMixin:
    """Mixin which adds support for s3 configuration.

    This mixin will upload to proper path by using key generator from config.

    """

    def __init__(
        self,
        # This arg is none because Django tries to init field with no args.
        s3_config: configs.S3FileTypeConfig | None = None,
        validate_key_pattern: bool = True,
        verbose_name: str | None = None,
        **kwargs,
    ) -> None:
        self.s3_config = s3_config
        self.validate_key_pattern = validate_key_pattern
        super().__init__(
            verbose_name=verbose_name,  # type: ignore
            **kwargs,
        )

    @cached_property
    def validators(
        self,
    ) -> collections.abc.Sequence[
        collections.abc.Callable[[files.FieldFile | str], None],
    ]:
        """Get validators."""
        validators = super().validators  # type: ignore
        validators.append(self._validate_file_existence)
        if self.validate_key_pattern:
            validators.append(self._validate_key)
        return validators

    def generate_filename(
        self,
        instance: models.Model,
        filename: str,
    ) -> str:
        """Generate filename via config."""
        if not self.s3_config:
            raise exceptions.ImproperlyConfigured(  # pragma: no cover
                "Please set s3_config for field",
            )
        filename = self.s3_config.key(filename=filename)
        filename = utils.validate_file_name(  # type: ignore
            filename,
            allow_relative_path=True,
        )
        return self.storage.generate_filename(filename)  # type: ignore

    def _validate_file_existence(
        self,
        value: files.FieldFile | str,
    ) -> None:
        """Check key is present in bucket."""
        is_field_file = isinstance(value, files.FieldFile)
        file_obj = getattr(value, "_file", None)
        if is_field_file and file_obj:
            # Skip if value has reference to actual file
            return  # pragma: no cover

        try:
            if not self.storage.exists(str(value)):  # type: ignore
                raise exceptions.ValidationError(
                    _("File does not exist."),
                )
        # If url comes not from s3, then botocore on key validation will
        # raise ParamValidationError
        except botocore.exceptions.ParamValidationError as error:
            raise exceptions.ValidationError(
                error,
            ) from error  # pragma: no cover
        except exceptions.SuspiciousFileOperation as error:
            raise exceptions.ValidationError(
                error,
            ) from error  # pragma: no cover

    def _validate_key(
        self,
        value: files.FieldFile | str,
    ) -> None:
        """Check that key valid for s3 config."""
        is_field_file = isinstance(value, files.FieldFile)
        file_obj = getattr(value, "_file", None)
        if is_field_file and file_obj:
            # Skip if value has reference to actual file
            # It will be saved with key generated by config
            return  # pragma: no cover

        if not self.s3_config:
            raise exceptions.ImproperlyConfigured(  # pragma: no cover
                "Please set s3_config for field",
            )

        if not self.s3_config.key.validate(str(value)):
            raise exceptions.ValidationError(
                _("File's path doesn't match with config's pattern."),
            )


class S3FileField(S3FileFieldMixin, models.FileField):
    """FileField with S3 capabilities."""


class S3ImageField(S3FileFieldMixin, models.ImageField):
    """FileField with S3 capabilities."""



import collections.abc
import dataclasses
import warnings

import boto3
import botocore.client
import botocore.config
import botocore.credentials
import botocore.exceptions
import botocore.response
import mypy_boto3_s3
import mypy_boto3_s3.type_defs

from . import configs

AccessKeyGetter = collections.abc.Callable[
    [],
    botocore.credentials.Credentials,
]
S3EndpointUrlGetter = collections.abc.Callable[
    [],
    str | None,
]
RegionGetter = collections.abc.Callable[
    [],
    str,
]


def get_boto3_session(
    access_key_getter: AccessKeyGetter,
    region: RegionGetter | str = "",
) -> boto3.session.Session:
    """Get AWS session."""
    if callable(region):
        region = region()  # pragma: no cover
    credentials = access_key_getter()
    return boto3.session.Session(
        aws_session_token=credentials.token or None,
        aws_access_key_id=credentials.access_key or None,
        aws_secret_access_key=credentials.secret_key or None,
        region_name=region,
    )


def get_boto3_s3_client(
    session: boto3.session.Session | None = None,
    access_key_getter: AccessKeyGetter | None = None,
    region: RegionGetter | str = "",
    s3_endpoint_url_getter: S3EndpointUrlGetter | None = None,
    config: botocore.config.Config | None = None,
) -> mypy_boto3_s3.S3Client:
    """Prepare boto3's s3 client for usage."""
    if access_key_getter:
        session = get_boto3_session(
            access_key_getter=access_key_getter,
            region=region,
        )
    if not session:
        raise ValueError(
            "Please pass either session or access_key_getter",
        )  # pragma: no cover

    endpoint_url = None
    if s3_endpoint_url_getter:
        endpoint_url = s3_endpoint_url_getter()

    return session.client(
        service_name="s3",  # type: ignore
        endpoint_url=endpoint_url,
        config=config,
    )


def get_boto3_s3_resource(
    session: boto3.session.Session | None = None,
    access_key_getter: AccessKeyGetter | None = None,
    region: RegionGetter | str = "",
    s3_endpoint_url_getter: S3EndpointUrlGetter | None = None,
    config: botocore.config.Config | None = None,
) -> mypy_boto3_s3.S3ServiceResource:
    """Prepare boto3's s3 resource for usage."""
    if access_key_getter:
        session = get_boto3_session(  # pragma: no cover
            access_key_getter=access_key_getter,
            region=region,
        )
    if not session:
        raise ValueError(
            "Please pass either session or access_key_getter",
        )  # pragma: no cover

    endpoint_url = None
    if s3_endpoint_url_getter:
        endpoint_url = s3_endpoint_url_getter()

    return session.resource(
        service_name="s3",  # type: ignore
        endpoint_url=endpoint_url,
        config=config,
    )


@dataclasses.dataclass
class S3UploadParams:
    """Representation of s3 upload params."""

    url: str
    params: dict[str, str]


class S3Client:
    """Client for interacting with s3 based on boto3 client."""

    def __init__(
        self,
        boto3_client: mypy_boto3_s3.S3Client,
        default_bucket: str,
        default_download_expiration: int = 3600,
    ) -> None:
        self.boto3_client = boto3_client
        self.default_bucket = default_bucket
        self.default_download_expiration = default_download_expiration

    def _get_fields(
        self,
        config: configs.S3FileTypeConfig,
        content_type: str,
        meta_data: dict[str, str],
    ) -> dict[str, int | str]:
        """Prepare fields for s3 upload."""
        fields: dict[str, int | str] = {
            "success_action_status": config.success_action_status,
            "Content-Type": content_type,
        }
        fields.update(**meta_data)
        if config.content_disposition:
            fields["Content-Disposition"] = config.content_disposition
        return fields

    def _get_conditions(
        self,
        config: configs.S3FileTypeConfig,
        content_type: str,
        meta_data: dict[str, str],
    ) -> list[list[str | int] | dict[str, str | int]]:
        """Prepare conditions for s3 upload."""
        conditions: list[list[str | int] | dict[str, str | int]] = [
            {"success_action_status": str(config.success_action_status)},
            {"Content-Type": content_type},
        ]
        if config.content_length_range:
            conditions.append(
                [
                    "content-length-range",
                    *list(config.content_length_range),
                ],
            )
        if config.content_disposition:
            conditions.append(
                {"Content-Disposition": config.content_disposition},
            )
        for key, value in meta_data.items():
            conditions.append({key: value})
        return conditions

    def prepare_meta_data(
        self,
        config: configs.S3FileTypeConfig,
        extra_metadata: dict[str, str] | None = None,
    ) -> dict[str, str]:
        """Prepare metadata."""
        meta_data = {
            "x-amz-meta-config-name": config.name,
        }
        for key, value in (extra_metadata or {}).items():
            meta_data[f"x-amz-meta-{key}"] = value
        return meta_data

    def generate_params(
        self,
        filename: str,
        config: configs.S3FileTypeConfig,
        content_type: str,
        bucket: str = "",
        upload_folder: str = "",
        extra_metadata: dict[str, str] | None = None,
    ) -> S3UploadParams:
        """Generate params for s3 upload."""
        meta_data = self.prepare_meta_data(
            config=config,
            extra_metadata=extra_metadata,
        )
        for key in meta_data:
            if "_" in key:
                warnings.warn(
                    "Use `-` instead of `_` as separator for key. "
                    f"Example {key.replace('x-amz-meta-', '')} -> "
                    f"{key.replace('x-amz-meta-', '').replace('_', '-')}.",
                    stacklevel=2,
                )
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/generate_presigned_post.html
        s3_params = self.boto3_client.generate_presigned_post(
            Bucket=bucket or self.default_bucket,
            Key="/".join(
                filter(None, (upload_folder, config.key(filename=filename))),
            ),
            Fields=self._get_fields(
                config=config,
                content_type=content_type,
                meta_data=meta_data,
            ),
            Conditions=self._get_conditions(
                config=config,
                content_type=content_type,
                meta_data=meta_data,
            ),
            ExpiresIn=config.expires_in,
        )
        return S3UploadParams(
            url=s3_params["url"],
            params=s3_params["fields"],
        )

    def upload_file(
        self,
        filename: str,
        config: configs.S3FileTypeConfig,
        file_obj: mypy_boto3_s3.type_defs.FileobjTypeDef,
        bucket: str = "",
    ) -> str:
        """Upload file to s3."""
        key = config.key(filename=filename)
        self.boto3_client.upload_fileobj(
            Fileobj=file_obj,
            Bucket=bucket or self.default_bucket,
            Key=key,
        )
        return key

    def download_file(
        self,
        key: str,
        file_obj: mypy_boto3_s3.type_defs.FileobjTypeDef,
        bucket: str = "",
    ) -> mypy_boto3_s3.type_defs.FileobjTypeDef:
        """Download file from s3."""
        self.boto3_client.download_fileobj(
            Fileobj=file_obj,
            Bucket=bucket or self.default_bucket,
            Key=key,
        )
        return file_obj

    def generate_presigned_url(
        self,
        key: str,
        bucket: str = "",
        expiration: int = 0,
    ) -> str:
        """Generate url for viewing/downloading file."""
        return self.boto3_client.generate_presigned_url(
            ClientMethod="get_object",
            Params={
                "Bucket": bucket or self.default_bucket,
                "Key": key,
            },
            ExpiresIn=expiration or self.default_download_expiration,
        )

    def get_file_metadata(
        self,
        key: str,
        bucket: str = "",
    ) -> mypy_boto3_s3.type_defs.HeadObjectOutputTypeDef:
        """Get file's metadata."""
        return self.boto3_client.head_object(
            Key=key,
            Bucket=bucket or self.default_bucket,
        )

    def is_file_in_bucket(
        self,
        key: str,
        bucket: str = "",
    ) -> bool:
        """Check if file is in bucket."""
        try:
            self.get_file_metadata(
                key=key,
                bucket=bucket,
            )
            return True
        except botocore.exceptions.ClientError as error:
            if error.response.get("Error", {}).get("Code") == "404":
                return False
            raise error  # pragma: no cover

    def copy_object(
        self,
        key: str,
        source_key: str,
        bucket: str = "",
        source_bucket: str = "",
    ) -> None:
        """Copy file object from copy source to key path."""
        self.boto3_client.copy_object(
            Bucket=bucket or self.default_bucket,
            CopySource=f"{source_bucket or self.default_bucket}/{source_key}",
            Key=key,
        )

    def delete_object(
        self,
        key: str,
        bucket: str = "",
    ) -> None:
        """Delete file object from s3 bucket."""
        self.boto3_client.delete_object(
            Bucket=bucket or self.default_bucket,
            Key=key,
        )



from rest_framework import routers

from . import views

router = routers.DefaultRouter()
router.register(
    prefix="",
    viewset=views.S3GetParamsView,
    basename="s3",
)
urlpatterns = router.urls



s3v4_signature_fields = (
    "success_action_status",
    "Content-Disposition",
    "x-amz-meta-user-id",
    "x-amz-meta-config-name",
    "Content-Type",
    "key",
    "x-amz-security-token",
    "x-amz-algorithm",
    "x-amz-credential",
    "x-amz-date",
    "policy",
    "x-amz-signature",
)



import typing
import urllib.parse

from django.core import validators
from django.core.files.storage import Storage, default_storage
from rest_framework import fields, serializers
from rest_framework.utils.formatting import lazy_format

from .. import configs


class S3FileTypeConfigField(serializers.ChoiceField):
    """Custom Choice field for s3 configs.

    Represent config choice in api and covert this choice to proper
    S3FileTypeConfig instance.

    """

    def __init__(self, **kwargs) -> None:
        super().__init__(choices=(), **kwargs)

    def _get_choices(self) -> dict[str, str]:
        """Get choices from S3FileTypeConfig."""
        current_choices = tuple(
            (
                config_name,
                config_name,
            )
            for config_name in configs.S3FileTypeConfig.configs
        )
        choices = super()._get_choices()
        if current_choices != choices:
            self._set_choices(current_choices)
        return super()._get_choices()

    def _set_choices(self, choices: tuple[tuple[str, str], ...]) -> None:
        """Update choices.

        Redefined to avoid recursion.

        """
        self.grouped_choices = fields.to_choices_dict(choices)
        self._choices = fields.flatten_choices_dict(
            self.grouped_choices,
        )

        # Map the string representation of choices to the underlying value.
        # Allows us to deal with eg. integer choices while supporting either
        # integer or string input, but still get the correct datatype out.
        self.choice_strings_to_values = {
            str(key): key for key in self._choices
        }

    choices = property(_get_choices, _set_choices)

    def to_internal_value(
        self,
        data: typing.Any,
    ) -> configs.S3FileTypeConfig | None:
        """Convert api data to S3FileTypeConfig."""
        try:
            return configs.S3FileTypeConfig.configs[str(data)]
        except KeyError:
            self.fail("invalid_choice", input=data)


class S3UploadURLField(serializers.CharField):
    """Char serializer field for S3 `files or keys or URLs or objects`.

    Convert url/path to a valid s3 key for bucket.

    Example:
    -------
    https://www.saritasa-s3-tools.s3.localhost/some-folder/file.txt will
    become `some-folder/file.txt` which is location of file in bucket (key).
    Same result will apply if AWS_LOCATION is used.
    https://www.saritasa-s3-tools.s3.localhost/locations/some-folder/file.txt
    will become `some-folder/file.txt`.

    """

    def __init__(self, **kwargs) -> None:
        """Make custom initialization."""
        # Remove explicit max_length limit in field
        # It causes issues with auto spec generation and validation.
        # Since it can return full urls with auth query, which can easily pass
        # a limit specified in model field(which is by default 100), it causes
        # confusion for openapi specs validators.
        max_length = kwargs.pop("max_length", None)
        self.storage: Storage = kwargs.pop("storage", default_storage)
        self.max_length_validator = None
        super().__init__(**kwargs)
        if max_length is not None:
            # Manually validate max length so that api specs would generate
            # without limit
            self.max_length_validator = validators.MaxLengthValidator(
                limit_value=max_length,
                message=lazy_format(
                    self.error_messages["max_length"],
                    max_length=max_length,
                ),
            )
        # Append this validator to enable invalid code for spec
        validator_for_spec = validators.MinLengthValidator(
            limit_value=0,
            message=self.error_messages["invalid"],
        )
        validator_for_spec.code = "invalid"
        self.validators.append(validator_for_spec)

    def to_internal_value(self, data: typing.Any) -> str:
        """Validate `data` and convert it to internal value.

        Cut domain from url to save it in file field.

        """
        if not isinstance(data, str):
            self.fail("invalid")  # pragma: no cover
        # Crop server domain and port and get relative path to avatar
        file_url = urllib.parse.urlparse(url=data).path

        # Crop S3 bucket name
        file_url = file_url.split(
            f"{self.storage.bucket_name}/",  # type: ignore
        )[-1].lstrip("/")

        # Normalize url
        file_url = urllib.parse.unquote_plus(file_url)

        # Remove aws-location prefix to keep only file name as key
        aws_location = self.storage.location  # type: ignore
        if aws_location and file_url.startswith(aws_location):
            file_url = file_url.split(f"{aws_location}/")[-1]
        if self.max_length_validator:
            self.max_length_validator(file_url)
        return file_url

    def to_representation(self, value: typing.Any) -> str | None:
        """Return full file url."""
        if not value:
            return None
        if isinstance(value, str):
            return self.storage.url(name=value)
        return value.url
